"""Flask application that creates a RESTful API from SQLAlchemy models."""

from flask import Blueprint, jsonify

from flask_sandboy.service import Service
from flask_sandboy.models import SerializableModel
from flask_sandboy.exception import (BadRequestException,
        ForbiddenException,
        NotAcceptableException,
        NotFoundException,
        ConflictException,
        ServerErrorException,
        NotImplementedException,
        ServiceUnavailableException,
        )

__version__ = '0.0.2.5'

class Sandboy(object):
    def __init__(self, app, db, models):
        self.app = app
        self.db = db
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['sandboy'] = self
        self.init_app(app, models)

    def init_app(self, app, models):
        self.blueprint = Blueprint('sandboy', __name__)
        @self.blueprint.errorhandler(BadRequestException)
        @self.blueprint.errorhandler(ForbiddenException)
        @self.blueprint.errorhandler(NotAcceptableException)
        @self.blueprint.errorhandler(NotFoundException)
        @self.blueprint.errorhandler(ConflictException)
        @self.blueprint.errorhandler(ServerErrorException)
        @self.blueprint.errorhandler(NotImplementedException)
        @self.blueprint.errorhandler(ServiceUnavailableException)
        def handle_application_error(error):
            """Handler used to send JSON error messages rather than default HTML
            ones."""
            response = jsonify(error.to_dict())
            response.status_code = error.code
            return response

        self.register(models)
        app.register_blueprint(self.blueprint)

    def register(self, cls_list):
        """Register a class to be given a REST API."""
        for cls in cls_list:
            serializable_model = type(
                    cls.__name__ + 'Serializable',
                    (cls, SerializableModel),
                    {})
            new_endpoint = type(
                    cls.__name__ + 'Endpoint',
                    (Service,),
                    {'__model__': serializable_model,
                    '__db__': self.db})
            view_func = new_endpoint.as_view(new_endpoint.__model__.__tablename__)
            self.blueprint.add_url_rule(
                    '/' + new_endpoint.__model__.__tablename__,
                    defaults={'resource_id': None},
                    view_func=view_func)
            self.blueprint.add_url_rule(
                    '/{resource}/<resource_id>'.format(
                        resource=new_endpoint.__model__.__tablename__),
                view_func=view_func)



from flask import request, jsonify, make_response
from flask.views import MethodView

from flask_sandboy.models import verify_fields
from flask_sandboy.exception import NotFoundException

class Service(MethodView):
    """Base class for all resources."""

    __model__ = None
    __db__ = None

    def get(self, resource_id):
        """Return response to HTTP GET request."""
        if resource_id is None:
            return self.all_resources()
        else:
            resource = self.resource(resource_id)
            if not resource:
                raise NotFoundException
            return jsonify(resource.to_dict())

    def all_resources(self):
        """Return all resources of this type as a JSON list."""
        if not 'page' in request.args:
            resources = self.__db__.session.query(self.__model__).all()
        else:
            resources = self.__model__.query.paginate(int(request.args['page'])).items
        return jsonify({'resources': [resource.to_dict() for resource in resources]})

    @verify_fields
    def post(self, resource_id=None):
        """Return response to HTTP POST request."""
        resource = self.__model__.query.filter_by(**request.json).first()
        # resource already exists; don't create it again
        if resource:
            return self._no_content_response()
        instance = self.__model__(**request.json)
        self.__db__.session.add(instance)
        self.__db__.session.commit()
        return self._created_response(instance.to_dict())

    def delete(self, resource_id):
        """Return response to HTTP DELETE request."""
        instance = self.resource(resource_id)
        self.__db__.session.delete(instance)
        self.__db__.session.commit()
        return self._no_content_response()

    @verify_fields
    def put(self, resource_id):
        """Return response to HTTP PUT request."""
        instance = self.resource(resource_id)
        if instance is None:
            instance = self.__model__(**request.json)
        else:
            instance.from_dict(request.json)
        self.__db__.session.add(instance)
        self.__db__.session.commit()
        return self._created_response(instance.to_dict())

    @verify_fields
    def patch(self, resource_id):
        """Return response to HTTP PATCH request."""
        resource = self.resource(resource_id)
        resource.from_dict(request.json)
        self.__db__.session.add(resource)
        self.__db__.session.commit()
        return self._created_response(resource.to_dict())

    def resource(self, resource_id):
        """Return resource represented by this *resource_id*."""
        resource = self.__db__.session.query(self.__model__).get(resource_id)
        if not resource:
            return None
        return resource

    @staticmethod
    def _no_content_response():
        """Return an HTTP 204 "No Content" response."""
        response = make_response()
        response.status_code = 204
        return response

    @staticmethod
    def _created_response(resource):
        """Return an HTTP 201 "Created" response."""
        response = jsonify(resource)
        response.status_code = 201
        return response



"""JSON-based Exception classes that mimic today's error messages sent by
DMF."""

class EndpointException(Exception):
    """Base class for all Exceptions."""

    def __init__(self, message=None, payload=None):
        super(EndpointException, self).__init__(message)
        self.message = message
        self.payload = payload

    def to_dict(self):
        """Return a dictionary representation of the exception."""
        as_dict = dict(self.payload or ())
        as_dict['message'] = self.message
        return as_dict


class BadRequestException(EndpointException):
    """Raised when a request contains illegal arguments, is missing arguments,
    can't be decoded properly, or the request is trying to do something that
    doesn't make sense."""

    code = 400


class ForbiddenException(EndpointException):
    """Raised when a request asks us to do something that we won't do because it
    violates the application logic. *Does not refer to an authentication failure.*
    Rather, it means the action requested is forbidden by the application."""

    code = 403


class NotFoundException(EndpointException):
    """Raised when the endpoint (or a resource it refers to) is not found. Can
    also be used if a resource referred to in a request (e.g. a specific job in
    a /job_status request) can not be found."""

    code = 404


class NotAcceptableException(EndpointException):
    """Raised when the client does not Accept any of the Content-Types we are
    capable of generating."""

    code = 406


class ConflictException(EndpointException):
    """Similar to a ServerErrorException (HTTP 500) but there is some action the
    client may take to resolve the conflict, after which the request can be
    resubmitted. A request to reprocess a job not marked for reprocessing, for
    example, could cause this exception to be raised."""

    code = 409


class ServerErrorException(EndpointException):
    """Raised when the application itself encounters an error not related to the
    request itself (for example, a database error)."""

    code = 500


class NotImplementedException(EndpointException):
    """Raised when the application does not implement the functionality being
    requested. Note that this doesn't refer to an HTTP method not being
    implemented for a given endpoint (which would be a 405 error)."""

    code = 501


class ServiceUnavailableException(EndpointException):
    """Raised when a resource is temporarily unavailable (e.g. not being able to
    get a database connection). Setting the *Retry-After* header gives the
    length of the delay, if it is known. Otherwise, this is treated as a 500
    error."""

    code = 503



import datetime
from functools import wraps

from flask import request

from flask_sandboy.exception import ForbiddenException, BadRequestException

class SerializableModel(object):
    """A SQLAlchemy model mixin class that can serialize itself as JSON."""

    def to_dict(self):
        """Return dict representation of class by iterating over database
        columns."""
        value = {}
        for column in self.__table__.columns:
            attribute = getattr(self, column.name)
            if isinstance(attribute, datetime.datetime):
                attribute = str(attribute)
            value[column.name] = attribute
        return value

    def from_dict(self, attributes):
        """Update the current instance based on attribute->value items in
        *attributes*."""
        for attribute in attributes:
            setattr(self, attribute, attributes[attribute])
        return self

def verify_fields(f):
    """A decorator to automatically extract required and optional fields from json data, making them
    attributes of the class instance."""
    @wraps(f)
    def decorated(instance, *args, **kwargs):
        """The decorator function."""
        data = request.get_json(force=True, silent=True)
        if not data:
            raise BadRequestException("No data received from request")
        for required in instance.__model__.__table__.columns:
            if required.name in instance.__model__.__table__.primary_key.columns:
                continue
            if required.name not in data:
                raise ForbiddenException('{} required'.format(required))
        return f(instance, *args, **kwargs)

    return decorated



from __future__ import print_function
from distutils.core import setup, Command
from setuptools import setup, find_packages
import codecs
import os
import sys
import re

class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys,subprocess
        errno = subprocess.call([
            sys.executable,
            'runtests.py',
            '--strict',
            '--verbose',
            '--tb=long',
            'tests'])
        raise SystemExit(errno)

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

long_description = read('README.rst')

setup(
    name='flask_sandboy',
    version=find_version('flask_sandboy', '__init__.py'),
    url='http://github.com/jeffknupp/flask_sandboy/',
    license='BSD License',
    author='Jeff Knupp',
    install_requires=['Flask>=0.10.1', ],
    tests_require=['pytest'],
    cmdclass = {'test': PyTest},
    author_email='jeff@jeffknupp.com',
    description='Automated REST APIs for SQLAlchemy models',
    long_description=long_description,
    packages=['flask_sandboy'],
    include_package_data=True,
    platforms='any',
    zip_safe=False,
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        ],
    extras_require={
        'testing': ['pytest'],
      }
)
